﻿module coneneok.location;

interface _Location
{
	char[] toString();
	void fromString(char[] a);
	uint clickLength();
	void addChoice(uint choices);
	void opPostInc();
	bit opEquals(_Location a);
	static uint nextChoice(_Location target, _Location current);
	char[] choicesString();
}

template Location_()
{
	private import std.string;
	private import std.conv;
	private uint _clickLength = 0;
	uint clickLength() { return _clickLength; }
	private char[] _choicesString = "";
	char[] choicesString() { return _choicesString; }
	void opPostInc() { _clickLength++; }
	
	char[] toString()
	{
		return _choicesString ~ "_" ~ std.string.toString(_clickLength);
	}
	
	void fromString(char[] a)
	{
		uint _ = find(a, "_");
		_choicesString = a[0.._];
		_clickLength = toUint(a[_ + 1..$]);
	}
	
	void addChoice(uint choices)
	{
		_choicesString ~= std.string.toString(choices);
		_clickLength = 0;
	}
	
	bit opEquals(_Location a) { return toString() == a.toString(); }
	
	static uint nextChoice(_Location target, _Location current)
	{
		char[] c0 = target.choicesString;
		char[] c1 = current.choicesString;
		
		if (c0.length <= c1.length) throw new Error("nextChoice");
		if (0 != c0.find(c1)) throw new Error("nextChoice");
		//dout.writef("%d : %d", c1.length, c0[c1.length]);
		return toUint("" ~ c0[c1.length]);
	}
}

class Location : _Location
{
	unittest
	{
		_Location l = new Location();
		assert("_0" == l.toString());
		
		l.fromString("_11");
		assert(11 == l.clickLength);
		
		l.addChoice(2);
		assert("2_0" == l.toString());
		
		l++;
		assert("2_1" == l.toString());
		
		l.addChoice(3);
		assert("23_0" == l.toString());
		
		l++;
		assert(new Location("23_1") == l);
		
		assert(4 == nextChoice(new Location("2345_0"), l));
	}
	
	mixin Location_;
	
	this() {}
	this(char[] a) { fromString(a); }
}
