﻿module hdk.network;

class Network
{
	unittest
	{
		Network net = new Network();
		
		Node a = new Node();
		Node b = new Node();
		
		assert(!net.has(a));
		assert(!net.has(b));
		
		net ~= a;
		net ~= b;
		
		assert(net.has(a));
		assert(net.has(b));
		
		
		assert(!net.parentOf(a));
		assert(!net.parentOf(b));
		
		net.link(a, b);
		
		assert(!net.parentOf(a));
		assert(a == net.parentOf(b));
		
		net.cut(a, b);
		
		assert(!net.parentOf(a));
		assert(!net.parentOf(b));
		
		
		assert(a == net.rootOf(a));
		assert(b == net.rootOf(b));
		
		net.link(a, b);
		
		assert(a == net.rootOf(a));
		assert(a == net.rootOf(b));
		
		net.clear(a);
		
		assert(b == net.rootOf(b));
		
		
		net ~= a;
		net.link(b, a);
		
		assert(1 == net.childrenOf(b).length);
		assert(a == net.childrenOf(b)[0]);
	}
	
	private Node[] nodeList;
	private Link[] linkList;
	
	bit has(Node a)
	{
		foreach (Node node; nodeList)
		{
			if (node == a) return true;
		}
		return false;
	}
	
	void opCatAssign(Node a)
	{
		nodeList ~= a;
	}
	
	Node parentOf(Node a)
	{
		foreach (Link link; linkList)
		{
			if (link.second == a) return link.first;
		}
		return null;
	}
	
	void link(Node a, Node b)
	{
		linkList ~= new Link(a, b);
	}
	
	void cut(Node a, Node b)
	{
		Link[] newLinkList;
		Link ab = new Link(a, b);
		foreach (Link link; linkList)
		{
			if (ab != link) newLinkList ~= link;
		}
		linkList = newLinkList;
	}
	
	Node rootOf(Node n)
	{
		Node result = n;
		Node current = n;
		while (true)
		{
			current = parentOf(current);
			if (!current) break;
			result = current;
		}
		return result;
	}
	
	int opApply(int delegate(inout Node) dg)
	{
		foreach (Node a; nodeList)
		{
			int result = dg(a);
			if (result) return result;
		}
		return 0;
	}
	
	void clear(Node n)
	{
		Node[] newNodeList;
		foreach (Node a; nodeList)
		{
			if (n != a) newNodeList ~= a;
		}
		nodeList = newNodeList;
		
		Link[] newLinkList;
		foreach (Link a; linkList)
		{
			if (!a.has(n)) newLinkList ~= a;
		}
		linkList = newLinkList;
	}
	
	Node[] childrenOf(Node n)
	{
		Node[] result;
		foreach (Link a; linkList)
		{
			if (a.first == n) result ~= a.second;
		}
		return result;
	}
}

class Node {}

private class Link
{
	Node first, second;
	
	this(Node f, Node s)
	{
		first = f;
		second = s;
	}
	
	bit opEquals(Link a)
	{
		return first == a.first && second == a.second ? true : false;
	}
	
	bit has(Node a)
	{
		return first == a || second == a ? true : false;
	}
}
