﻿module hdk.sdlttf;

void setTtfFileName(dchar[] a) { SdlTtfThread.getInstance().ttfFileName = a; } ///

void setTtfFontStyleBoldItalicUnderline(bit bold, bit italic, bit underline) ///
{
	SdlTtfThread.getInstance().bold = bold;
	SdlTtfThread.getInstance().italic = italic;
	SdlTtfThread.getInstance().underline = underline;
}

/// 32bit rgbaで512x32の大きさにtextを分割する
dchar[][] splitTextTexture512x32(dchar[] text)
{
	return SdlTtfThread.getInstance().splitTextTexture512x32(text);
}

/// テキストとフォントから32bit rgbaで512x32の大きさのイメージを作る
uint[] createTextTexture512x32(dchar[] text)
{
	return SdlTtfThread.getInstance().createTextTexture512x32(text);
}


version (hdk_lib) {
private import sdl_ttf;
private import std.string;
private import std.c.stdlib;
private import std.utf;
private import hdk.error;
private import hdk.sdlimage;
private import std.cstream;
private import std.thread;
private import hdk.sdlbase;

private class SdlTtfThread : Thread
{
	private static SdlTtfThread _instance;
	static this() { _instance = new SdlTtfThread(); }
	static ~this() { delete _instance; }
	static SdlTtfThread getInstance() { return _instance; }
	
	dchar[] ttfFileName = `C:\WINDOWS\Fonts\msgothic.ttc`;
	bit bold, italic, underline;
	
	private this()
	{
		start();
		yield();
	}
	
	private bit loop = true;
	
	~this()
	{
		loop = false;
		while (getState() != TS.TERMINATED) { yield(); }
	}
	
	int run()
	{
		try
		{
			auto SdlTtf sdlTtf = new SdlTtf();
			while (loop)
			{
				if (splitingText && !resultOfSpliting)
				{
					char[][] list = sdlTtf.splitTextTexture512x32(toUTF8(splitingText),
						toUTF8(ttfFileName), bold, italic, underline);
					foreach (char[] a; list) resultOfSpliting ~= toUTF32(a);
					splitingText = null;
				}
				if (creatingText && !resultOfCreating)
				{
					resultOfCreating = sdlTtf.createTextTexture512x32(toUTF8(creatingText),
						toUTF8(ttfFileName), bold, italic, underline);
					creatingText = null;
				}
				yield();
				sleep(1);
			}
		}
		catch (Object e)
		{
			derr.writeLine("SdlTtfThread:" ~ e.toString());
		}
		return 0;
	}
	
	private dchar[] splitingText;
	private dchar[][] resultOfSpliting;
	
	synchronized dchar[][] splitTextTexture512x32(dchar[] text)
	{
		splitingText = text;
		resultOfSpliting = null;
		while (true)
		{
			if (!splitingText && resultOfSpliting) break;
			yield();
		}
		return resultOfSpliting;
	}
	
	private dchar[] creatingText;
	private uint[] resultOfCreating;
	
	synchronized uint[] createTextTexture512x32(dchar[] text)
	{
		creatingText = text;
		resultOfCreating = null;
		while (true)
		{
			if (!creatingText && resultOfCreating) break;
			yield();
		}
		return resultOfCreating;
	}
}

private class SdlTtf
{
	mixin SdlError;
	
	private static bit initialized = false;
	private SdlTtfFontMap fontMap;
	
	this()
	{
		if (initialized) error(__LINE__, "initialized");
		TTF_Init();
		fontMap = new SdlTtfFontMap();
		initialized = true;
	}
	
	~this()
	{
		//delete fontMap; // 削除できない?
		TTF_Quit();
		initialized = false;
	}
	
	char[][] splitTextTexture512x32(char[] text,
		char[] ttfFileName, bit bold, bit italic, bit underline)
	{
		TTF_Font* font = fontMap.getFont(ttfFileName, bold, italic, underline);
		char[][] result;
		char[] left, right;
		left = text.dup;
		
		while (left.length > 0)
		{
			assert(right.length == 0);
			
			while (true)
			{
				if (getWidth(font, left) <= 512) break;
				char a = left[left.length - 1];
				left.length = left.length - 1;
				
				right = a ~ right;
			}
			
			result ~= left.dup;
			left = right.dup;
			right.length = 0;
		}
		
		return result;
	}
	
	uint[] createTextTexture512x32(char[] text,
		char[] ttfFileName, bit bold, bit italic, bit underline)
	out (result)
	{
		assert(512 * 32 == result.length);
	}
	body
	{
		TTF_Font* font = fontMap.getFont(ttfFileName, bold, italic, underline);
		
		int width, height;
		getSize(font, text, width, height);
		if (width > 512 || height > 32) error(__LINE__);
		
		uint[] textImage;
		try
		{
			textImage = createTextImage(font, text);
			assert(width * height == textImage.length);
		}
		catch (Object e)
		{
			dout.writeLine(e.toString());
			textImage.length = width * height;
			textImage[] = uint.max;
		}
		
		uint[] result = new uint[512 * 32];
		for (int i = 0; i < height; i++)
		{
			memcpy(&result[512 * i], &textImage[width * i], uint.sizeof * width);
		}
		
		return result;
	}
	
	private uint[] createTextImage(TTF_Font* font, char[] text)
	{
		SDL_Color foreColor;
		foreColor.r = foreColor.g = foreColor.b = foreColor.unused = 255;
		
		SDL_Surface* textSurface;
		try
		{
			textSurface = TTF_RenderUTF8_Blended(font, toStringz(text), foreColor);
		}
		catch (Object e)
		{
			// 時々作成に失敗するらしい、原因はよくわからない
			error(__LINE__, format(`TTF_RenderUTF8_Blended "%s"`, text));
		}
		if (!textSurface) error(__LINE__, format(`!textSurface "%s"`, text));
		
		uint[] result = SdlImage.readRgba32Pixels(textSurface);
		SDL_FreeSurface(textSurface);
		return result;
	}
	
	private static uint getWidth(TTF_Font* font, char[] text)
	{
		int width, height;
		getSize(font, text, width, height);
		assert(height <= 32);
		return width;
	}
	
	private static void getSize(
		TTF_Font* font, char[] text, out int width, out int height)
	{
		if (-1 == TTF_SizeUTF8(font, toStringz(text), &width, &height))
		{
			throw new Error("TTF_SizeUTF8:" ~ toString(SDL_GetError()));
		}
	}
	
	private void clearFontMap()
	{
		delete fontMap;
		fontMap = new SdlTtfFontMap();
	}
}

private class SdlTtfFontMap
{
	private TTF_Font*[char[]] map;
	
	~this()
	{
		foreach (TTF_Font* a; map) TTF_CloseFont(a);
	}
	
	TTF_Font* getFont(char[] ttfFileName, bit bold, bit italic, bit underline)
	{
		if (!(ttfFileName in map))
		{
			map[ttfFileName] = createHeight32Font(ttfFileName);
		}
		setTextStyle(map[ttfFileName], bold, italic, underline);
		return map[ttfFileName];
	}
	
	private static TTF_Font* createHeight32Font(char[] ttfFileName)
	out (result)
	{
		assert(TTF_FontHeight(result) <= 32);
	}
	body
	{
		for (int i = 32 - 1; i >= 1; i--)
		{
			TTF_Font* result = TTF_OpenFont(toStringz(ttfFileName), i);
			if (!result) throw new Error("TTF_OpenFont");
			if (TTF_FontHeight(result) <= 32) return result;
			TTF_CloseFont(result);
		}
		throw new Error("createHeight32Font");
	}
	
	private static void setTextStyle(TTF_Font* font, bit bold, bit italic, bit underline)
	{
		int style = 0;
		if (bold) style |= TTF_STYLE_BOLD;
		if (italic) style |= TTF_STYLE_ITALIC;
		if (underline) style |= TTF_STYLE_UNDERLINE;
		TTF_SetFontStyle(font, style);
	}
}

} // version (hdk_lib)
