module hdk.wbcommands; 

template DrawCommands()
{

	void projection(Matrix m)
	{
		class ProjectionCommand : Command
		{
			private final WindowResource wr;
			private final Matrix m;

			this(WindowResource wr, Matrix m)
			{
				this.wr = wr;
				this.m = m;
			}

			void opCall()
			{
				wr.projection(m);
			}
		}

		nextDrawCommandList ~= new ProjectionCommand(base, m);
	}

	void modelView(Matrix m)
	{
		class ModelViewCommand : Command
		{
			private final WindowResource wr;
			private final Matrix m;

			this(WindowResource wr, Matrix m)
			{
				this.wr = wr;
				this.m = m;
			}

			void opCall()
			{
				wr.modelView(m);
			}
		}

		nextDrawCommandList ~= new ModelViewCommand(base, m);
	}

	void depthTestWriting(bit test, bit writing)
	{
		class DepthTestWritingCommand : Command
		{
			private final WindowResource wr;
			private final bit test;
			private final bit writing;

			this(WindowResource wr, bit test, bit writing)
			{
				this.wr = wr;
				this.test = test;
				this.writing = writing;
			}

			void opCall()
			{
				wr.depthTestWriting(test, writing);
			}
		}

		nextDrawCommandList ~= new DepthTestWritingCommand(base, test, writing);
	}

	void cullingEnableFront(bit enable, bit front)
	{
		class CullingEnableFrontCommand : Command
		{
			private final WindowResource wr;
			private final bit enable;
			private final bit front;

			this(WindowResource wr, bit enable, bit front)
			{
				this.wr = wr;
				this.enable = enable;
				this.front = front;
			}

			void opCall()
			{
				wr.cullingEnableFront(enable, front);
			}
		}

		nextDrawCommandList ~= new CullingEnableFrontCommand(base, enable, front);
	}

	void setShaderFloat(dchar[] shader, char[] name, float a)
	{
		class SetShaderFloatCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] shader;
			private final char[] name;
			private final float a;

			this(WindowResource wr, dchar[] shader, char[] name, float a)
			{
				this.wr = wr;
				this.shader = shader;
				this.name = name;
				this.a = a;
			}

			void opCall()
			{
				wr.setShaderFloat(shader, name, a);
			}
		}

		nextDrawCommandList ~= new SetShaderFloatCommand(base, shader, name, a);
	}

	void setShaderInt(dchar[] shader, char[] name, int a)
	{
		class SetShaderIntCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] shader;
			private final char[] name;
			private final int a;

			this(WindowResource wr, dchar[] shader, char[] name, int a)
			{
				this.wr = wr;
				this.shader = shader;
				this.name = name;
				this.a = a;
			}

			void opCall()
			{
				wr.setShaderInt(shader, name, a);
			}
		}

		nextDrawCommandList ~= new SetShaderIntCommand(base, shader, name, a);
	}

	void setShaderVector(dchar[] shader, char[] name, Vector a)
	{
		class SetShaderVectorCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] shader;
			private final char[] name;
			private final Vector a;

			this(WindowResource wr, dchar[] shader, char[] name, Vector a)
			{
				this.wr = wr;
				this.shader = shader;
				this.name = name;
				this.a = a;
			}

			void opCall()
			{
				wr.setShaderVector(shader, name, a);
			}
		}

		nextDrawCommandList ~= new SetShaderVectorCommand(base, shader, name, a);
	}

	void setShaderMatrixArray(dchar[] shader, char[] name, Matrix[] ma)
	{
		class SetShaderMatrixArrayCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] shader;
			private final char[] name;
			private final Matrix[] ma;

			this(WindowResource wr, dchar[] shader, char[] name, Matrix[] ma)
			{
				this.wr = wr;
				this.shader = shader;
				this.name = name;
				this.ma = ma;
			}

			void opCall()
			{
				wr.setShaderMatrixArray(shader, name, ma);
			}
		}

		nextDrawCommandList ~= new SetShaderMatrixArrayCommand(base, shader, name, ma);
	}

	void drawPcnt(dchar[] shader, dchar[] vertexBuffer, dchar[] texture0, dchar[] texture1 = null, dchar[] texture2 = null, dchar[] texture3 = null)
	{
		class DrawPcntCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] shader;
			private final dchar[] vertexBuffer;
			private final dchar[] texture0;
			private final dchar[] texture1;
			private final dchar[] texture2;
			private final dchar[] texture3;

			this(WindowResource wr, dchar[] shader, dchar[] vertexBuffer, dchar[] texture0, dchar[] texture1 = null, dchar[] texture2 = null, dchar[] texture3 = null)
			{
				this.wr = wr;
				this.shader = shader;
				this.vertexBuffer = vertexBuffer;
				this.texture0 = texture0;
				this.texture1 = texture1;
				this.texture2 = texture2;
				this.texture3 = texture3;
			}

			void opCall()
			{
				wr.drawPcnt(shader, vertexBuffer, texture0, texture1, texture2, texture3);
			}
		}

		nextDrawCommandList ~= new DrawPcntCommand(base, shader, vertexBuffer, texture0, texture1, texture2, texture3);
	}

	void drawPcntmw(dchar[] shader, char[] matrixIndex, char[] weight, dchar[] vertexBuffer, dchar[] texture0, dchar[] texture1 = null, dchar[] texture2 = null, dchar[] texture3 = null)
	{
		class DrawPcntmwCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] shader;
			private final char[] matrixIndex;
			private final char[] weight;
			private final dchar[] vertexBuffer;
			private final dchar[] texture0;
			private final dchar[] texture1;
			private final dchar[] texture2;
			private final dchar[] texture3;

			this(WindowResource wr, dchar[] shader, char[] matrixIndex, char[] weight, dchar[] vertexBuffer, dchar[] texture0, dchar[] texture1 = null, dchar[] texture2 = null, dchar[] texture3 = null)
			{
				this.wr = wr;
				this.shader = shader;
				this.matrixIndex = matrixIndex;
				this.weight = weight;
				this.vertexBuffer = vertexBuffer;
				this.texture0 = texture0;
				this.texture1 = texture1;
				this.texture2 = texture2;
				this.texture3 = texture3;
			}

			void opCall()
			{
				wr.drawPcntmw(shader, matrixIndex, weight, vertexBuffer, texture0, texture1, texture2, texture3);
			}
		}

		nextDrawCommandList ~= new DrawPcntmwCommand(base, shader, matrixIndex, weight, vertexBuffer, texture0, texture1, texture2, texture3);
	}

	void drawRectangle(Vector color)
	{
		class DrawRectangleCommand : Command
		{
			private final WindowResource wr;
			private final Vector color;

			this(WindowResource wr, Vector color)
			{
				this.wr = wr;
				this.color = color;
			}

			void opCall()
			{
				wr.drawRectangle(color);
			}
		}

		nextDrawCommandList ~= new DrawRectangleCommand(base, color);
	}

	void drawTexture(dchar[] texture, Vector color)
	{
		class DrawTextureCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] texture;
			private final Vector color;

			this(WindowResource wr, dchar[] texture, Vector color)
			{
				this.wr = wr;
				this.texture = texture;
				this.color = color;
			}

			void opCall()
			{
				wr.drawTexture(texture, color);
			}
		}

		nextDrawCommandList ~= new DrawTextureCommand(base, texture, color);
	}

	void drawText(dchar[] text, Vector color)
	{
		class DrawTextCommand : Command
		{
			private final WindowResource wr;
			private final dchar[] text;
			private final Vector color;

			this(WindowResource wr, dchar[] text, Vector color)
			{
				this.wr = wr;
				this.text = text;
				this.color = color;
			}

			void opCall()
			{
				wr.drawText(text, color);
			}
		}

		nextDrawCommandList ~= new DrawTextCommand(base, text, color);
	}
}

template Decoration()
{

	uint width()
	{
		return base.width();
	}

	uint height()
	{
		return base.height();
	}

	float mouseX()
	{
		return base.mouseX();
	}

	float mouseY()
	{
		return base.mouseY();
	}
}
