﻿module hdk.window;
version (hdk_lib)
{
	private import hdk.wrlayers;
	private import hdk.sdlttf;
}
private import std.thread;
private import hdk.error;
private import std.openrj;
private import std.file;
private import std.conv;
private import std.cstream;
private import std.gc;
private import hdk.control;
import hdk.wr;
import hdk.math;
import hdk.sdlbase;
import hdk.error;

auto class Window : Thread, WindowResource
{
	interface Command { void opCall(); }
	
	private import hdk.wbcommands;
	mixin DrawCommands;
	mixin Decoration;
	mixin HdkError_;
	
	int opApply(int delegate(inout uint) dg) { return base.opApply(dg); }
	void inputHandler(void delegate(uint input) dg) { return base.inputHandler(dg); }
	
	private bit endFlag = false;
	private WrLayers _base;
	
	WrLayers base()
	{
		if (!_base) error(__LINE__, "base is null");
		return _base;
	}
	
	this()
	{
		initializeTotalResourceSizeLimit();
		start();
		while (!_base)
		{
			yield();
			if (getState() == TS.TERMINATED) error(__LINE__, "initialize");
		}
	}
	
	private const char[] INIT_RJ = "init.rj";
	private bit resourceOverflow = false;
	private final uint totalResourceSizeLimit;
	
	private void initializeTotalResourceSizeLimit()
	{
		Database data = new Database(cast(char[])read(INIT_RJ), 0);
		totalResourceSizeLimit = toUint(data[0]["totalResourceSizeLimit"].replace("_", ""));
	}
	
	~this()
	{
		endFlag = true;
		if (getState() == TS.RUNNING) wait();
	}
	
	private uint preTicks;
	
	int run()
	{
		try
		{
			auto WrLayer7 wnd = new WrLayer7();
			_base = wnd;
			base.doEvents();
			preTicks = getTicks();
			while (!endFlag)
			{
				if (resourceOverflow)
				{
					dout.writeString("resource clearing...");
					wnd.clearResource();
					resourceOverflow = false;
					dout.writeLine("ok");
					dout.flush();
				}
				
				foreach (Command a; drawCommandList) a();
				executeTerminableCommands();
				
				if (loadingBackBuffer)
				{
					dout.writeString("back buffer reading...");
					delete backBuffer;
					backBuffer = base.readBackBuffer();
					dout.writeLine("ok");
					dout.flush();
					fullCollect();
					loadingBackBuffer = false;
				}
				
				tickRootControl();
				
				if (!base.doEvents()) error(__LINE__, "closed");
				
				sleep(1);
				while (getTicks() - preTicks < 1000 / 60) sleep(1);
				preTicks = getTicks();
			}
		}
		catch (Exception e)
		{
			dout.writeLine("hdk.window.Window thread : " ~ e.toString());
			dout.flush();
		}
		finally
		{
			if (_rootControl) _rootControl.setWindowResource(null);
			_base = null;
		}
		return 0;
	}
	
	private Command[] drawCommandList, nextDrawCommandList;
	
	void update()
	{
		drawCommandList = null;
		
		if (base.totalResourceSize > totalResourceSizeLimit) resourceOverflow = true;
		
		while (nextDrawCommandList.length != 0)
		{
			drawCommandList ~= nextDrawCommandList[0];
			nextDrawCommandList = nextDrawCommandList[1..$];
		}
		sleep(1);
	}
	
	private TerminableCommand[] terminableCommandList;
	
	void postTerminableCommand(TerminableCommand cmd)
	{
		terminableCommandList ~= cmd;
	}
	
	private void executeTerminableCommands()
	{
		TerminableCommand[] temp;
		foreach (TerminableCommand a; terminableCommandList)
		{
			if (a.alive) temp ~= a;
		}
		terminableCommandList = temp;
		
		foreach (TerminableCommand a; terminableCommandList) a(base);
	}
	
	private uint[] backBuffer;
	private bit loadingBackBuffer = false;
	
	uint[] readBackBuffer()
	{
		loadingBackBuffer = true;
		while (base && loadingBackBuffer) { sleep(1); }
		return backBuffer.dup;
	}
	
	
	private final RootControl _rootControl;
	
	Control rootControl()
	{
		if (!_rootControl)
		{
			_rootControl = new RootControl(base);
			inputHandler(&_rootControl.input);
		}
		return _rootControl;
	}
	
	private void tickRootControl()
	{
		if (!_rootControl) return;
		_rootControl.draw();
		_rootControl.clearOtherTree();
	}
	
	bit alive()
	{
		return _base ? true : false;
	}
}
