﻿module hdk.wr;
private import hdk.math;

/// 有効期限付きのコマンド、aliveがfalseになるまでopCallが呼ばれる
interface TerminableCommand
{
	void opCall(WindowResource wr); ///
	bit alive(); ///
}

/// hdk.Windowの定義
interface WindowResource
{
	void update(); ///
	
	void projection(Matrix m); ///
	void modelView(Matrix m); ///
	
	void depthTestWriting(bit test, bit writing); /// updateでtrue, trueに再設定される
	void cullingEnableFront(bit enable, bit front); ///
	
	void setShaderFloat(dchar[] shader, char[] name, float a); ///
	void setShaderInt(dchar[] shader, char[] name, int a); /// multi texture indexもこれで
	void setShaderVector(dchar[] shader, char[] name, Vector a); ///
	void setShaderMatrixArray(dchar[] shader, char[] name, Matrix[] ma); ///
	
	void drawPcnt(
		dchar[] shader, dchar[] vertexBuffer,
		dchar[] texture0, dchar[] texture1 = null, dchar[] texture2 = null, dchar[] texture3 = null
		); ///
	void drawPcntmw(
		dchar[] shader, char[] matrixIndex, char[] weight, dchar[] vertexBuffer,
		dchar[] texture0, dchar[] texture1 = null, dchar[] texture2 = null, dchar[] texture3 = null
		); ///
	
	void drawRectangle(Vector color); ///
	void drawTexture(dchar[] texture, Vector color); ///
	void drawText(dchar[] text, Vector color); ///
	
	uint width(); ///
	uint height(); ///
	int opApply(int delegate(inout uint) dg); /// 入力
	void inputHandler(void delegate(uint input) dg); ///
	float mouseX(); /// 左から右へ[-1.0, 1.0]
	float mouseY(); /// 下から上へ[-1.0, 1.0]
	
	void postTerminableCommand(TerminableCommand cmd); /// 最後に描画される
	
	uint[] readBackBuffer(); /// width * height rgba
}

interface WrLayers : WindowResource
{
	bit doEvents();
	uint totalResourceSize();
	void clearResource();
}
