﻿import std.cstream;
import hdk.window;
import std.path;
import std.file;
import nanami.toutf8fromsjis;
import std.utf;
import nanami.model;
import nanami.camera;
import std.math;
import hdk.control;
import std.string;
import nanami.png;
import std.gc;

int main(char[][] args)
{
	try
	{
		if (args.length < 2) throw new Error("file name not found");
		foreach (inout char[] a; args) a = toUtf8FromSjis(a);
		isfile(args[1]);
		chdir(getDirName(args[0]));
		
		auto Window wnd = new Window();
		dout.writeLine("window initialized"); dout.flush();
		
		dout.writeLine("model loading..."); dout.flush();
		Model model = new Model(toUTF32(args[1]), true);
		dout.writeLine("ok"); dout.flush();
		
		// "smesh_toon.vert"のmatrixArrayの長さと同じ
		const uint BONE_LENGTH_LIMIT = 34;
		if (model.boneLength > BONE_LENGTH_LIMIT)
		{
			dout.writefln("boneLength=%d >= LIMIT", model.boneLength);
			dout.writeLine("continue?(y/n)");
			if ("y" != din.readLine()) throw new Error("bone length limit over");
		}
		
		model.projection = Matrix.perspectiveFov;
		wnd.postTerminableCommand(model);
		
		uint motionLength = model.motionLength;
		uint length1 = model.height >= 2 ? model.width(1) : 0;
		uint length2_ = model.height >= 3 ? model.height - 2 : 0;
		Menu menu = new Menu(wnd.rootControl, motionLength, length1, length2_);
		
		CameraControl camera = new CameraControl(wnd.rootControl);
		
		uint[] morphIteratorList = new uint[length2_];
		uint preTick = getTicks();
		
		MotionIterator mi = model.motionIterator;
		
		float f = 0.0;
		while (true)
		{
			f += 0.01;
			model.lightDirection(normalize(vector(cos(f), -1, sin(f))));
			model.modelView = camera.modelView;
			
			for (int i = 0; i < menu.visibleItemLength; i++)
			{
				model[1, i].visible = menu.visibleItem(i);
			}
			
			bit tick = false;
			if (getTicks() - preTick >= 100)
			{
				preTick = getTicks();
				tick = true;
			}
			
			for (int i = 0; i < menu.morphItemLength; i++)
			{
				if (!menu.activeMorph(i)) continue;
				if (!tick) continue;
				
				model[2 + i, morphIteratorList[i]].visible = false;
				
				morphIteratorList[i]++;
				if (morphIteratorList[i] >= model.width(2 + i))
				{
					morphIteratorList[i] = 0;
				}
				
				model[2 + i, morphIteratorList[i]].visible = true;
			}
			
			if (model.hasMotion)
			{
				mi.y = menu.motionIndex;
				mi++;
				if (mi.isRangeOut) mi.x = 0;
			}
			
			bit screenShot = false;
			foreach (uint input; wnd)
			{
				if (input == 10) screenShot = true;
			}
			
			wnd.update();
			
			if (screenShot)
			{
				const char[] DIR = "screenshot";
				if (!exists(DIR)) mkdir(DIR);
				
				if (!model.hasMotion)
				{
					uint[] backBuffer = wnd.readBackBuffer();
					writePng(DIR ~ `\sc.png`, backBuffer, wnd.width, wnd.height);
					delete backBuffer;
					dout.writefln(`created "%s\sc.png"`, DIR);
				}
				else
				{
					mi.x = 0;
					while (!mi.isRangeOut)
					{
						char[] fileName = format(`%s\sc%04d.png`, DIR, mi.x);
						uint[] backBuffer = wnd.readBackBuffer();
						writePng(fileName, backBuffer, wnd.width, wnd.height);
						delete backBuffer;
						dout.writefln(`created "%s"`, fileName);
						dout.flush();
						fullCollect();
						sleep(10);
						mi.x = mi.x + 4;
					}
					mi.x = 0;
				}
			}
			
			sleep(15);
		}
	}
	catch (HdkError e)
	{
		dout.writeLine(e.toString());
	}
	catch (Exception e)
	{
		dout.flush();
		derr.writeLine(e.toString());
		dout.writeLine("---- failed ----");
		din.readLine();
		dout.writeLine("end of process");
	}
	return 0;
}

class CameraControl : Control
{
	private Camera camera;
	
	this(Control parent)
	{
		super(parent);
		camera = new Camera();
	}
	
	override bit inner(Point p)
	{
		return true;
	}
	
	override void input(uint id)
	{
		switch (id)
		{
			case 0: camera.pushLeft(); break;
			case 1: camera.popLeft(); break;
			case 2: camera.pushRight(); break;
			case 3: camera.popRight(); break;
			default: break;
		}
	}
	
	override void draw()
	{
		camera.tick(wr.mouseX, wr.mouseY);
	}
	
	Matrix modelView()
	{
		return camera.getViewMatrix();
	}
}

class Menu : Control
{
	private CheckBox option;
	private RadioGroup motion;
	private CheckBoxGroup check, morph;
	
	private Vector foreColor() { return Vector.create(0, 0, 0, 1); }
	const float width = 1.8, height = 0.08;
	
	this(Control parent, uint motionLength, uint checkLength, uint morphLength)
	{
		super(parent);
		
		option = new CheckBox(this);
		option.foreColor = foreColor;
		with (option)
		{
			text = "option";
			rectangle = Rectangle(-1.0, 1.0 - height, 0.25, height);
			backColor = Vector(1, 1, 1, 0.4);
			checkedEvent = &changeVisible;
		}
		
		if (motionLength != 0)
		{
			motion = new RadioGroup(this);
			motion.foreColor = foreColor;
			with (motion)
			{
				length = motionLength;
				for (int i = 0; i < length; i++) motion[i].text = toUTF32(.toString(i));
				rectangle = Rectangle(-0.9, 0.8, width, height);
				backColor = Vector(1, 0, 0, 0.6);
				borderColor = Vector(1, 0, 0, 0.1);
				selectedEvent = &selectMotion;
			}
		}
		
		if (checkLength != 0)
		{
			check = new CheckBoxGroup(this);
			check.foreColor = foreColor;
			with (check)
			{
				length = checkLength;
				for (int i = 0; i < length; i++) check[i].text = toUTF32(.toString(i));
				rectangle = Rectangle(-0.9, 0.7, width, height);
				backColor = Vector(0, 1, 0, 0.6);
				borderColor = Vector(0, 1, 0, 0.1);
			}
		}
		
		if (morphLength != 0)
		{
			morph = new CheckBoxGroup(this);
			morph.foreColor = foreColor;
			with (morph)
			{
				length = morphLength;
				for (int i = 0; i < length; i++) morph[i].text = toUTF32(.toString(i));
				rectangle = Rectangle(-0.9, 0.6, width, height);
				backColor = Vector(0, 0, 1, 0.6);
				borderColor = Vector(0, 0, 1, 0.1);
			}
		}
		
		changeVisible(false);
	}
	
	uint motionIndex = 0;
	void selectMotion(uint a) { motionIndex = a; }
	
	void changeVisible(bit a)
	{
		if (motion) motion.visible = a;
		if (check) check.visible = a;
		if (morph) morph.visible = a;
	}
	
	// [1][index]
	bit visibleItem(uint index) { return check[index].value; }
	
	// [index][t]
	bit activeMorph(uint index) { return morph[index].value; }
	
	uint visibleItemLength() { return check ? check.length : 0; }
	uint morphItemLength() { return morph ? morph.length : 0; }
}
